/* vim:ts=8:sts=4:sw=4:noai:noexpandtab
 * 
 * Rate regulation.
 *
 * Copyright (c) 2006-2010 Miru Limited.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#if !defined (__PGM_IMPL_FRAMEWORK_H_INSIDE__) && !defined (PGM_COMPILATION)
#	error "Only <framework.h> can be included directly."
#endif

#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#	pragma once
#endif
#ifndef __PGM_IMPL_RATE_CONTROL_H__
#define __PGM_IMPL_RATE_CONTROL_H__

typedef struct pgm_rate_t pgm_rate_t;

#include <pgm/types.h>
#include <pgm/time.h>
#include <impl/thread.h>

PGM_BEGIN_DECLS

struct pgm_rate_t {
	ssize_t		rate_per_sec;
	ssize_t		rate_per_msec;
	size_t		iphdr_len;

	ssize_t		rate_limit;		/* signed for math */
	pgm_time_t	last_rate_check;
	pgm_spinlock_t	spinlock;
};

PGM_GNUC_INTERNAL void pgm_rate_create (pgm_rate_t*, const ssize_t, const size_t, const uint16_t);
PGM_GNUC_INTERNAL void pgm_rate_destroy (pgm_rate_t*);
PGM_GNUC_INTERNAL bool pgm_rate_check2 (pgm_rate_t*, pgm_rate_t*, const size_t, const bool);
PGM_GNUC_INTERNAL bool pgm_rate_check (pgm_rate_t*, const size_t, const bool);
PGM_GNUC_INTERNAL pgm_time_t pgm_rate_remaining2 (pgm_rate_t*, pgm_rate_t*, const size_t);
PGM_GNUC_INTERNAL pgm_time_t pgm_rate_remaining (pgm_rate_t*, const size_t);

PGM_END_DECLS

#endif /* __PGM_IMPL_RATE_CONTROL_H__ */
