/*
 * Note: this file originally auto-generated by mib2c using
 *  : mib2c.column_enums.conf,v 5.2 2003/02/22 04:09:25 hardaker Exp $
 */

#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#	pragma once
#endif
#ifndef PGMMIB_ENUMS_H
#define PGMMIB_ENUMS_H

/* enums for column pgmNeIfPgmEnable */
           #define PGMNEIFPGMENABLE_ENABLE		1
           #define PGMNEIFPGMENABLE_DISABLE		2

/* enums for column pgmNeTsiStateBits */
           #define PGMNETSISTATEBITS_INITIALISING		0
           #define PGMNETSISTATEBITS_SPMSQNSTATEVALID		1
           #define PGMNETSISTATEBITS_DLRCANPROVIDEPARITY		2

/* enums for column pgmNeTsiRtxSequenceNumberType */
           #define PGMNETSIRTXSEQUENCENUMBERTYPE_SELECTIVE		1
           #define PGMNETSIRTXSEQUENCENUMBERTYPE_TG		2

/* enums for column pgmNeTsiRtxStateBits */
           #define PGMNETSIRTXSTATEBITS_INITIALISING		0
           #define PGMNETSIRTXSTATEBITS_ELIMINATING		1
           #define PGMNETSIRTXSTATEBITS_REDIRECTING		2
           #define PGMNETSIRTXSTATEBITS_STATECREATEDBYNULLNAK		3
           #define PGMNETSIRTXSTATEBITS_LISTNAKENTRY		4
           #define PGMNETSIRTXSTATEBITS_PARITYSTATE		5

/* enums for column pgmNeTsiPollType */
           #define PGMNETSIPOLLTYPE_GENERAL		1
           #define PGMNETSIPOLLTYPE_DLR		2

/* enums for column pgmSourceAdvMode */
           #define PGMSOURCEADVMODE_DATA		1
           #define PGMSOURCEADVMODE_TIME		2
           #define PGMSOURCEADVMODE_APPLCTRL		3
           #define PGMSOURCEADVMODE_OTHER		4

/* enums for column pgmSourceLateJoin */
           #define PGMSOURCELATEJOIN_ENABLE		1
           #define PGMSOURCELATEJOIN_DISABLE		2

/* enums for column pgmSourceFEC */
           #define PGMSOURCEFEC_DISABLED		1
           #define PGMSOURCEFEC_ENABLEDFIXEDPACKETSIZE		2
           #define PGMSOURCEFEC_ENABLEDVARIABLEPACKETSIZE		3

/* enums for column pgmReceiverSendNaks */
           #define PGMRECEIVERSENDNAKS_ENABLED		1
           #define PGMRECEIVERSENDNAKS_DISABLED		2

/* enums for column pgmReceiverLateJoin */
           #define PGMRECEIVERLATEJOIN_ENABLED		1
           #define PGMRECEIVERLATEJOIN_DISABLED		2

/* enums for column pgmReceiverDeliveryOrder */
           #define PGMRECEIVERDELIVERYORDER_UNORDERED		1
           #define PGMRECEIVERDELIVERYORDER_ORDERED		2

/* enums for column pgmReceiverMcastNaks */
           #define PGMRECEIVERMCASTNAKS_ENABLED		1
           #define PGMRECEIVERMCASTNAKS_DISABLED		2

#endif /* PGMMIB_ENUMS_H */
