/*
 * Note: this file originally auto-generated by mib2c using
 *  : mib2c.column_defines.conf,v 5.1 2002/05/08 05:42:47 hardaker Exp $
 */

#if defined(_MSC_VER) && (_MSC_VER >= 1200)
#	pragma once
#endif
#ifndef PGMMIB_COLUMNS_H
#define PGMMIB_COLUMNS_H

/* column number definitions for table pgmNeIfConfigTable */
       #define COLUMN_PGMNEIFCONFIGINDEX		1
       #define COLUMN_PGMNEIFPGMENABLE		2
       #define COLUMN_PGMNEIFNAKRPTINTERVAL		3
       #define COLUMN_PGMNEIFNAKRPTRATE		4
       #define COLUMN_PGMNEIFNAKRDATAINTERVAL		5
       #define COLUMN_PGMNEIFNAKELIMINATEINTERVAL		6

/* column number definitions for table pgmNeIfPerformanceTable */
       #define COLUMN_PGMNEIFPERFORMANCEINDEX		1
       #define COLUMN_PGMNEIFREXMITSTATES		2
       #define COLUMN_PGMNEIFREXMITTIMEDOUT		3
       #define COLUMN_PGMNEIFINSPMS		4
       #define COLUMN_PGMNEIFOUTSPMS		5
       #define COLUMN_PGMNEIFINPARITYSPMS		6
       #define COLUMN_PGMNEIFOUTPARITYSPMS		7
       #define COLUMN_PGMNEIFINRDATA		8
       #define COLUMN_PGMNEIFOUTRDATA		9
       #define COLUMN_PGMNEIFINPARITYRDATA		10
       #define COLUMN_PGMNEIFOUTPARITYRDATA		11
       #define COLUMN_PGMNEIFINRDATANOSESSIONERRORS		12
       #define COLUMN_PGMNEIFUNIQUENAKS		13
       #define COLUMN_PGMNEIFINNAKS		14
       #define COLUMN_PGMNEIFOUTNAKS		15
       #define COLUMN_PGMNEIFUNIQUEPARITYNAKS		16
       #define COLUMN_PGMNEIFINPARITYNAKS		17
       #define COLUMN_PGMNEIFOUTPARITYNAKS		18
       #define COLUMN_PGMNEIFINNAKNOSESSIONERRORS		19
       #define COLUMN_PGMNEIFINNAKSEQERRORS		20
       #define COLUMN_PGMNEIFINPARITYNAKTGERRORS		21
       #define COLUMN_PGMNEIFINNNAKS		22
       #define COLUMN_PGMNEIFOUTNNAKS		23
       #define COLUMN_PGMNEIFINPARITYNNAKS		24
       #define COLUMN_PGMNEIFOUTPARITYNNAKS		25
       #define COLUMN_PGMNEIFINNNAKNOSESSIONERRORS		26
       #define COLUMN_PGMNEIFINNCFS		27
       #define COLUMN_PGMNEIFOUTNCFS		28
       #define COLUMN_PGMNEIFINPARITYNCFS		29
       #define COLUMN_PGMNEIFOUTPARITYNCFS		30
       #define COLUMN_PGMNEIFINNCFNOSESSIONERRORS		31
       #define COLUMN_PGMNEIFINREDIRECTNCFS		32
       #define COLUMN_PGMNEIFMALFORMED		33
       #define COLUMN_PGMNEIFSPMFROMSOURCE		34
       #define COLUMN_PGMNEIFSPMBADSQN		35
       #define COLUMN_PGMNEIFSPMERROR		36
       #define COLUMN_PGMNEIFPOLLRANDOMIGNORE		37
       #define COLUMN_PGMNEIFPOLLTSISTATEERROR		38
       #define COLUMN_PGMNEIFPOLLPARENTERROR		39
       #define COLUMN_PGMNEIFPOLLTYPEERROR		40
       #define COLUMN_PGMNEIFPOLLERROR		41
       #define COLUMN_PGMNEIFPOLLSUCCESS		42
       #define COLUMN_PGMNEIFPOLLORIGINATED		43
       #define COLUMN_PGMNEIFPOLRNOSTATE		44
       #define COLUMN_PGMNEIFPOLRERROR		45
       #define COLUMN_PGMNEIFPOLRPARITYERROR		46
       #define COLUMN_PGMNEIFPOLRSUCCESS		47
       #define COLUMN_PGMNEIFPOLRORIGINATED		48
       #define COLUMN_PGMNEIFNCFERROR		49
       #define COLUMN_PGMNEIFNCFPARITYERROR		50
       #define COLUMN_PGMNEIFNCFPARTIALPARITY		51
       #define COLUMN_PGMNEIFNCFRECEIVED		52
       #define COLUMN_PGMNEIFNCFANTICIPATED		53
       #define COLUMN_PGMNEIFNCFREDIRECTING		54
       #define COLUMN_PGMNEIFNAKELIMINATED		55
       #define COLUMN_PGMNEIFNAKERROR		56
       #define COLUMN_PGMNEIFNAKPARITYERROR		57
       #define COLUMN_PGMNEIFNNAKELIMINATED		58
       #define COLUMN_PGMNEIFNNAKERROR		59
       #define COLUMN_PGMNEIFNNAKPARITYERROR		60
       #define COLUMN_PGMNEIFNNAKCONGESTIONREPORTS		61
       #define COLUMN_PGMNEIFNAKRETRYEXPIRED		62
       #define COLUMN_PGMNEIFNAKRETRYEXPIREDDLR		63
       #define COLUMN_PGMNEIFNAKFORWARDEDDLR		64
       #define COLUMN_PGMNEIFNAKRETRANSMITTED		65
       #define COLUMN_PGMNEIFRDATAELIMINATEDOIF		66
       #define COLUMN_PGMNEIFRDATAELIMINATEDSQN		67
       #define COLUMN_PGMNEIFINRDATAFRAGMENTS		68
       #define COLUMN_PGMNEIFRDATAFRAGMENTSNOSESSIONERRORS		69
       #define COLUMN_PGMNEIFRDATAFRAGMENTSELIMINATEDOIF		70
       #define COLUMN_PGMNEIFRDATAFRAGMENTSELIMINATEDSQN		71
       #define COLUMN_PGMNEIFOUTRDATAFRAGMENTS		72

/* column number definitions for table pgmNeTsiTable */
       #define COLUMN_PGMNETSIGLOBALID		1
       #define COLUMN_PGMNETSIDATASOURCEPORT		2
       #define COLUMN_PGMNETSISTATEBITS		3
       #define COLUMN_PGMNETSIDATADESTINATIONPORT		4
       #define COLUMN_PGMNETSISOURCEADDRESS		5
       #define COLUMN_PGMNETSIGROUPADDRESS		6
       #define COLUMN_PGMNETSIUPSTREAMADDRESS		7
       #define COLUMN_PGMNETSIUPSTREAMIFINDEX		8
       #define COLUMN_PGMNETSIDLRADDRESS		9

/* column number definitions for table pgmNeTsiPerformanceTable */
       #define COLUMN_PGMNETSIPERFORMANCEGLOBALID		1
       #define COLUMN_PGMNETSIPERFORMANCEDATASOURCEPORT		2
       #define COLUMN_PGMNETSISESSIONTRAILEDGESEQ		3
       #define COLUMN_PGMNETSISESSIONINCRSEQ		4
       #define COLUMN_PGMNETSILEADEDGESEQ		5
       #define COLUMN_PGMNETSIINSPMS		6
       #define COLUMN_PGMNETSIOUTSPMS		7
       #define COLUMN_PGMNETSIINPARITYSPMS		8
       #define COLUMN_PGMNETSIOUTPARITYSPMS		9
       #define COLUMN_PGMNETSITOTALREXMITSTATES		10
       #define COLUMN_PGMNETSITOTALREXMITTIMEDOUT		11
       #define COLUMN_PGMNETSIINRDATA		12
       #define COLUMN_PGMNETSIOUTRDATA		13
       #define COLUMN_PGMNETSIINPARITYRDATA		14
       #define COLUMN_PGMNETSIOUTPARITYRDATA		15
       #define COLUMN_PGMNETSIINRDATANOSTATEERRORS		16
       #define COLUMN_PGMNETSIUNIQUENAKS		17
       #define COLUMN_PGMNETSIINNAKS		18
       #define COLUMN_PGMNETSIOUTNAKS		19
       #define COLUMN_PGMNETSIUNIQUEPARITYNAKS		20
       #define COLUMN_PGMNETSIINPARITYNAKS		21
       #define COLUMN_PGMNETSIOUTPARITYNAKS		22
       #define COLUMN_PGMNETSIINNAKSEQERRORS		23
       #define COLUMN_PGMNETSIINNNAKS		24
       #define COLUMN_PGMNETSIOUTNNAKS		25
       #define COLUMN_PGMNETSIINPARITYNNAKS		26
       #define COLUMN_PGMNETSIOUTPARITYNNAKS		27
       #define COLUMN_PGMNETSIINNCFS		28
       #define COLUMN_PGMNETSIOUTNCFS		29
       #define COLUMN_PGMNETSIINPARITYNCFS		30
       #define COLUMN_PGMNETSIOUTPARITYNCFS		31
       #define COLUMN_PGMNETSISPMSEQUENCENUMBER		32
       #define COLUMN_PGMNETSITRANSMISSIONGROUPSIZE		33
       #define COLUMN_PGMNETSITIMEOUT		34
       #define COLUMN_PGMNETSILASTTTL		35
       #define COLUMN_PGMNETSILINKLOSSRATE		36
       #define COLUMN_PGMNETSIPATHLOSSRATE		37
       #define COLUMN_PGMNETSIRECEIVERLOSSRATE		38
       #define COLUMN_PGMNETSICONGESTIONREPORTLEAD		39
       #define COLUMN_PGMNETSICONGESTIONREPORTWORSTRECEIVER		40

/* column number definitions for table pgmNeTsiRtxTable */
       #define COLUMN_PGMNETSIRTXSEQUENCENUMBER		1
       #define COLUMN_PGMNETSIRTXSEQUENCENUMBERTYPE		2
       #define COLUMN_PGMNETSIRTXREQPARITYTGCOUNT		4
       #define COLUMN_PGMNETSIRTXTIMEOUT		5
       #define COLUMN_PGMNETSIRTXSTATEBITS		6

/* column number definitions for table pgmNeTsiRtxIfTable */
       #define COLUMN_PGMNETSIRTXIFINDEX		1
       #define COLUMN_PGMNETSIRTXIFPACKETCOUNT		2

/* column number definitions for table pgmNeTsiPolrTable */
       #define COLUMN_PGMNETSIPOLRSOURCE		1
       #define COLUMN_PGMNETSIPOLRSEQUENCENUMBER		2

/* column number definitions for table pgmNeTsiPollTable */
       #define COLUMN_PGMNETSIPOLLTYPE		1
       #define COLUMN_PGMNETSIPOLLSEQUENCE		2
       #define COLUMN_PGMNETSIPOLLCHILDBACKOFF		3
       #define COLUMN_PGMNETSIPOLLMASK		4
       #define COLUMN_PGMNETSIPOLLPERIOD		5
       #define COLUMN_PGMNETSIPOLLCOUNT		6
       #define COLUMN_PGMNETSIPOLLTIMEOUT		7

/* column number definitions for table pgmSourceTable */
       #define COLUMN_PGMSOURCEGLOBALID		1
       #define COLUMN_PGMSOURCESOURCEPORT		2
       #define COLUMN_PGMSOURCESOURCEADDRESS		3
       #define COLUMN_PGMSOURCEGROUPADDRESS		4
       #define COLUMN_PGMSOURCEDESTPORT		5
       #define COLUMN_PGMSOURCESOURCEGSI		6
       #define COLUMN_PGMSOURCESOURCEPORTNUMBER		7

/* column number definitions for table pgmSourceConfigTable */
       #define COLUMN_PGMSOURCECONFIGGLOBALID		1
       #define COLUMN_PGMSOURCECONFIGSOURCEPORT		2
       #define COLUMN_PGMSOURCETTL		3
       #define COLUMN_PGMSOURCEADVMODE		4
       #define COLUMN_PGMSOURCELATEJOIN		5
       #define COLUMN_PGMSOURCETXWMAXRTE		6
       #define COLUMN_PGMSOURCETXWSECS		7
       #define COLUMN_PGMSOURCETXWADVSECS		8
       #define COLUMN_PGMSOURCEADVIVL		9
       #define COLUMN_PGMSOURCESPMIVL		10
       #define COLUMN_PGMSOURCESPMHEARTBEATIVLMIN		11
       #define COLUMN_PGMSOURCESPMHEARTBEATIVLMAX		12
       #define COLUMN_PGMSOURCERDATABACKOFFIVL		13
       #define COLUMN_PGMSOURCEFEC		14
       #define COLUMN_PGMSOURCEFECTRANSMISSIONGRPSIZE		15
       #define COLUMN_PGMSOURCEFECPROACTIVEPARITYSIZE		16
       #define COLUMN_PGMSOURCESPMPATHADDRESS		17

/* column number definitions for table pgmSourcePerformanceTable */
       #define COLUMN_PGMSOURCEPERFORMANCEGLOBALID		1
       #define COLUMN_PGMSOURCEPERFORMANCESOURCEPORT		2
       #define COLUMN_PGMSOURCEDATABYTESSENT		3
       #define COLUMN_PGMSOURCEDATAMSGSSENT		4
       #define COLUMN_PGMSOURCEBYTESBUFFERED		5
       #define COLUMN_PGMSOURCEMSGSBUFFERED		6
       #define COLUMN_PGMSOURCEBYTESRETRANSMITTED		7
       #define COLUMN_PGMSOURCEMSGSRETRANSMITTED		8
       #define COLUMN_PGMSOURCEBYTESSENT		9
       #define COLUMN_PGMSOURCERAWNAKSRECEIVED		10
       #define COLUMN_PGMSOURCENAKSIGNORED		11
       #define COLUMN_PGMSOURCECKSUMERRORS		12
       #define COLUMN_PGMSOURCEMALFORMEDNAKS		13
       #define COLUMN_PGMSOURCEPACKETSDISCARDED		14
       #define COLUMN_PGMSOURCENAKSRCVD		15
       #define COLUMN_PGMSOURCEPARITYBYTESRETRANSMITTED		16
       #define COLUMN_PGMSOURCESELECTIVEBYTESRETRANSMITED		17
       #define COLUMN_PGMSOURCEPARITYMSGSRETRANSMITTED		18
       #define COLUMN_PGMSOURCESELECTIVEMSGSRETRANSMITTED		19
       #define COLUMN_PGMSOURCEBYTESADMIT		20
       #define COLUMN_PGMSOURCEMSGSADMIT		21
       #define COLUMN_PGMSOURCEPARITYNAKPACKETSRECEIVED		22
       #define COLUMN_PGMSOURCESELECTIVENAKPACKETSRECEIVED		23
       #define COLUMN_PGMSOURCEPARITYNAKSRECEIVED		24
       #define COLUMN_PGMSOURCESELECTIVENAKSRECEIVED		25
       #define COLUMN_PGMSOURCEPARITYNAKSIGNORED		26
       #define COLUMN_PGMSOURCESELECTIVENAKSIGNORED		27
       #define COLUMN_PGMSOURCEACKERRORS		28
       #define COLUMN_PGMSOURCEPGMCCACKER		29
       #define COLUMN_PGMSOURCETRANSMISSIONCURRENTRATE		30
       #define COLUMN_PGMSOURCEACKPACKETSRECEIVED		31
       #define COLUMN_PGMSOURCENNAKPACKETSRECEIVED		32
       #define COLUMN_PGMSOURCEPARITYNNAKPACKETSRECEIVED		33
       #define COLUMN_PGMSOURCESELECTIVENNAKPACKETSRECEIVED		34
       #define COLUMN_PGMSOURCENNAKSRECEIVED		35
       #define COLUMN_PGMSOURCEPARITYNNAKSRECEIVED		36
       #define COLUMN_PGMSOURCESELECTIVENNAKSRECEIVED		37
       #define COLUMN_PGMSOURCENNAKERRORS		38

/* column number definitions for table pgmReceiverTable */
       #define COLUMN_PGMRECEIVERGLOBALID		1
       #define COLUMN_PGMRECEIVERSOURCEPORT		2
       #define COLUMN_PGMRECEIVERINSTANCE		3
       #define COLUMN_PGMRECEIVERGROUPADDRESS		4
       #define COLUMN_PGMRECEIVERDESTPORT		5
       #define COLUMN_PGMRECEIVERSOURCEADDRESS		6
       #define COLUMN_PGMRECEIVERLASTHOP		7
       #define COLUMN_PGMRECEIVERSOURCEGSI		8
       #define COLUMN_PGMRECEIVERSOURCEPORTNUMBER		9
       #define COLUMN_PGMRECEIVERUNIQUEINSTANCE		10

/* column number definitions for table pgmReceiverConfigTable */
       #define COLUMN_PGMRECEIVERCONFIGGLOBALID		1
       #define COLUMN_PGMRECEIVERCONFIGSOURCEPORT		2
       #define COLUMN_PGMRECEIVERCONFIGINSTANCE		3
       #define COLUMN_PGMRECEIVERNAKBACKOFFIVL		4
       #define COLUMN_PGMRECEIVERNAKREPEATIVL		5
       #define COLUMN_PGMRECEIVERNAKNCFRETRIES		6
       #define COLUMN_PGMRECEIVERNAKRDATAIVL		7
       #define COLUMN_PGMRECEIVERNAKDATARETRIES		8
       #define COLUMN_PGMRECEIVERSENDNAKS		9
       #define COLUMN_PGMRECEIVERLATEJOIN		10
       #define COLUMN_PGMRECEIVERNAKTTL		11
       #define COLUMN_PGMRECEIVERDELIVERYORDER		12
       #define COLUMN_PGMRECEIVERMCASTNAKS		13
       #define COLUMN_PGMRECEIVERNAKFAILURETHRESHOLDTIMER		14
       #define COLUMN_PGMRECEIVERNAKFAILURETHRESHOLD		15

/* column number definitions for table pgmReceiverPerformanceTable */
       #define COLUMN_PGMRECEIVERPERFORMANCEGLOBALID		1
       #define COLUMN_PGMRECEIVERPERFORMANCESOURCEPORT		2
       #define COLUMN_PGMRECEIVERPERFORMANCEINSTANCE		3
       #define COLUMN_PGMRECEIVERDATABYTESRECEIVED		4
       #define COLUMN_PGMRECEIVERDATAMSGSRECEIVED		5
       #define COLUMN_PGMRECEIVERNAKSSENT		6
       #define COLUMN_PGMRECEIVERNAKSRETRANSMITTED		7
       #define COLUMN_PGMRECEIVERNAKFAILURES		8
       #define COLUMN_PGMRECEIVERBYTESRECEIVED		9
       #define COLUMN_PGMRECEIVERNAKSSUPPRESSED		10
       #define COLUMN_PGMRECEIVERCKSUMERRORS		11
       #define COLUMN_PGMRECEIVERMALFORMEDSPMS		12
       #define COLUMN_PGMRECEIVERMALFORMEDODATA		13
       #define COLUMN_PGMRECEIVERMALFORMEDRDATA		14
       #define COLUMN_PGMRECEIVERMALFORMEDNCFS		15
       #define COLUMN_PGMRECEIVERPACKETSDISCARDED		16
       #define COLUMN_PGMRECEIVERLOSSES		17
       #define COLUMN_PGMRECEIVERBYTESDELIVEREDTOAPP		18
       #define COLUMN_PGMRECEIVERMSGSDELIVEREDTOAPP		19
       #define COLUMN_PGMRECEIVERDUPSPMS		20
       #define COLUMN_PGMRECEIVERDUPDATAS		21
       #define COLUMN_PGMRECEIVERDUPPARITIES		22
       #define COLUMN_PGMRECEIVERNAKPACKETSSENT		23
       #define COLUMN_PGMRECEIVERPARITYNAKPACKETSSENT		24
       #define COLUMN_PGMRECEIVERSELECTIVENAKPACKETSSENT		25
       #define COLUMN_PGMRECEIVERPARITYNAKSSENT		26
       #define COLUMN_PGMRECEIVERSELECTIVENAKSSENT		27
       #define COLUMN_PGMRECEIVERPARITYNAKSRETRANSMITTED		28
       #define COLUMN_PGMRECEIVERSELECTIVENAKSRETRANSMITTED		29
       #define COLUMN_PGMRECEIVERNAKSFAILED		30
       #define COLUMN_PGMRECEIVERPARITYNAKSFAILED		31
       #define COLUMN_PGMRECEIVERSELECTIVENAKSFAILED		32
       #define COLUMN_PGMRECEIVERNAKSFAILEDRXWADVANCED		33
       #define COLUMN_PGMRECEIVERNAKSFALEDNCFRETRIESEXCEEDED		34
       #define COLUMN_PGMRECEIVERNAKSFAILEDDATARETRIESEXCEEDED		35
       #define COLUMN_PGMRECEIVERNAKSFAILEDGENEXPIRED		36
       #define COLUMN_PGMRECEIVERNAKFAILURESDELIVERED		37
       #define COLUMN_PGMRECEIVERPARITYNAKSSUPPRESSED		38
       #define COLUMN_PGMRECEIVERSELECTIVENAKSSUPPRESSED		39
       #define COLUMN_PGMRECEIVERNAKERRORS		40
       #define COLUMN_PGMRECEIVEROUTSTANDINGPARITYNAKS		41
       #define COLUMN_PGMRECEIVEROUTSTANDINGSELECTIVENAKS		42
       #define COLUMN_PGMRECEIVERLASTACTIVITY		43
       #define COLUMN_PGMRECEIVERNAKSVCTIMEMIN		44
       #define COLUMN_PGMRECEIVERNAKSVCTIMEMEAN		45
       #define COLUMN_PGMRECEIVERNAKSVCTIMEMAX		46
       #define COLUMN_PGMRECEIVERNAKFAILTIMEMIN		47
       #define COLUMN_PGMRECEIVERNAKFAILTIMEMEAN		48
       #define COLUMN_PGMRECEIVERNAKFAILTIMEMAX		49
       #define COLUMN_PGMRECEIVERNAKTRANSMITMIN		50
       #define COLUMN_PGMRECEIVERNAKTRANSMITMEAN		51
       #define COLUMN_PGMRECEIVERNAKTRANSMITMAX		52
       #define COLUMN_PGMRECEIVERACKSSENT		53
       #define COLUMN_PGMRECEIVERRXWTRAIL		54
       #define COLUMN_PGMRECEIVERRXWLEAD		55
       #define COLUMN_PGMRECEIVERNAKFAILURESLASTINTERVAL		56
       #define COLUMN_PGMRECEIVERLASTINTERVALNAKFAILURES		57

/* column number definitions for table pgmDlrSourceTable */
       #define COLUMN_PGMDLRSOURCEGLOBALID		1
       #define COLUMN_PGMDLRSOURCESOURCEPORT		2
       #define COLUMN_PGMDLRSOURCEGROUPADDRESS		3
       #define COLUMN_PGMDLRSOURCESOURCEGSI		4
       #define COLUMN_PGMDLRSOURCESOURCEPORTNUMBER		5

/* column number definitions for table pgmDlrSourceConfigTable */
       #define COLUMN_PGMDLRSOURCECONFIGGLOBALID		1
       #define COLUMN_PGMDLRSOURCECONFIGSOURCEPORT		2
       #define COLUMN_PGMDLRSOURCEGROUPTTL		3
       #define COLUMN_PGMDLRSOURCERDATABACKOFFIVL		4

/* column number definitions for table pgmDlrSourcePerformanceTable */
       #define COLUMN_PGMDLRSOURCEPERFORMANCEGLOBALID		1
       #define COLUMN_PGMDLRSOURCEPERFORMANCESOURCEPORT		2
       #define COLUMN_PGMDLRSOURCERDATAMSGSSENT		3
       #define COLUMN_PGMDLRSOURCERDATABYTESSENT		4
       #define COLUMN_PGMDLRSOURCEBYTESSENT		5
       #define COLUMN_PGMDLRSOURCENAKSRCVD		6
       #define COLUMN_PGMDLRSOURCENAKSIGNORED		7
       #define COLUMN_PGMDLRSOURCENAKERRORS		8
       #define COLUMN_PGMDLRSOURCEDISCARDS		9
       #define COLUMN_PGMDLRSOURCECKSUMERRORS		10
       #define COLUMN_PGMDLRSOURCENNAKSSENT		11
       #define COLUMN_PGMDLRSOURCEBYTESBUFFERED		12
       #define COLUMN_PGMDLRSOURCEMSGSBUFFERED		13
       #define COLUMN_PGMDLRSOURCEPARITYBYTESRETRANSMITTED		14
       #define COLUMN_PGMDLRSOURCESELECTIVEBYTESRETRANSMITED		15
       #define COLUMN_PGMDLRSOURCEPARITYMSGSRETRANSMITTED		16
       #define COLUMN_PGMDLRSOURCESELECTIVEMSGSRETRANSMITTED		17
       #define COLUMN_PGMDLRSOURCEBYTESADMIT		18
       #define COLUMN_PGMDLRSOURCEMSGSADMIT		19
       #define COLUMN_PGMDLRSOURCENAKPACKETSRECEIVED		20
       #define COLUMN_PGMDLRSOURCEPARITYNAKPACKETSRECEIVED		21
       #define COLUMN_PGMDLRSOURCESELECTIVENAKPACKETSRECEIVED		22
       #define COLUMN_PGMDLRSOURCEPARITYNAKSRECEIVED		23
       #define COLUMN_PGMDLRSOURCESELECTIVENAKSRECEIVED		24
       #define COLUMN_PGMDLRSOURCEPARITYNAKSIGNORED		25
       #define COLUMN_PGMDLRSOURCESELECTIVENAKSIGNORED		26
       #define COLUMN_PGMDLRSOURCEACKERRORS		27
       #define COLUMN_PGMDLRSOURCENNAKERRORS		28
       #define COLUMN_PGMDLRSOURCEACKPACKETSRECEIVED		29
       #define COLUMN_PGMDLRSOURCENNAKPACKETSRECEIVED		30
       #define COLUMN_PGMDLRSOURCEPARITYNNAKPACKETSRECEIVED		31
       #define COLUMN_PGMDLRSOURCESELECTIVENNAKPACKETSRECEIVED		32
       #define COLUMN_PGMDLRSOURCENNAKSRECEIVED		33
       #define COLUMN_PGMDLRSOURCEPARITYNNAKSRECEIVED		34
       #define COLUMN_PGMDLRSOURCESELECTIVENNAKSRECEIVED		35
#endif /* PGMMIB_COLUMNS_H */
