html {
	background-color: white;
	font-family: Verdana;
	font-size: 12px;
	color: black;
}

a, a:link, a:visited {
	color: #0033cc;
	text-decoration: none;
}

#header {
	text-align: right;
}

#header #hostname {
	font-weight: bold;
}

#header a {
	color: black;
}

#header a:hover {
	text-decoration: underline;
}

#footer {
	clear: both;
	margin-top: 3.5em;
	margin-bottom: 1em;
	padding-top: 20px;
	text-align: center;
}

#navigation a {
	color: black;
}

#navigation .tab {
	-moz-border-radius-topleft: 4px; -webkit-border-top-left-radius: 4px;
	-moz-border-radius-topright: 4px; -webkit-border-top-right-radius: 4px;
	padding: 4px 1em 2px;
	margin-right: 8px;
	float: left;
	font-weight: bold;
}

#navigation #tabtop,#tabline {
	background-color: #fb879c;
}

#navigation #tabbottom {
	background-color: #fbc1a9;
}

#navigation #tabline {
	clear: left;
	-moz-border-radius-topright: 4px; -webkit-border-top-right-radius: 4px;
	height: 4px;
}

#content {
	margin-top: 6px;
	padding: 3px;
}

#content a:hover {
	background: #ffffaa;
}

.heading {
	-moz-border-radius-topleft: 4px; -webkit-border-top-left-radius: 4px;
	-moz-border-radius-topright: 4px; -webkit-border-top-right-radius: 4px;
	background-color: #fb879c;
	padding: 6px;
	margin-bottom: 3px;
}

table {
	border-collapse: separate;
}

th {
	text-align: left;
}

#information {
	float: right;
}

.rounded {
	background-color: #fbc1a9;
	-moz-border-radius: 4px; -webkit-border-radius: 4px;
	padding: 5px;
	margin-top: 6px;
	margin-bottom: 6px;
	width: 25em;
}

.break {
	border-top: 3px solid white;
	margin-top: 1em;
	padding-top: 6px;
}

.bubbly {
	background-color: #fb879c;
	-moz-border-radius: 4px; -webkit-border-radius: 4px;
	padding: 4px;
}

.bubbly table {
	width: 100%;
}

.bubbly th,.bubbly td {
	border-bottom: 1px solid #bbbbbb;
}

.bubbly th {
	background-color: #fbc1a9;
	border-left: 1px solid #bbbbbb;
	padding: 2px 1px 2px 2px;
}

.bubbly td {
	background-color: white;
	padding: 4px;
}

.bubbly .empty {
	padding: 3em;
	text-align: center;
}
